/* Custom styles for Nicole documentation */

/* Increase spacing between page content and footer */
.md-content {
    margin-bottom: 3rem;  /* Adds 3rem (48px) of space at the bottom of content */
}

/* Reduce spacing between logo and site name */
.md-header__title {
    margin-left: 0 !important;  /* Reduce left margin (default is 1rem) */
    padding-left: 0 !important;
}
